<?php

namespace Drupal\entitylogic;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Extend this class for a basic EntityLogic implementation.
 */
class EntityLogicBase extends PluginBase implements EntityLogicInterface, ContainerFactoryPluginInterface {

  /**
   * The entity.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface
   */
  protected $entity;

  /**
   * Instance of EntityTypeManager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Holds the original entity values, from when the object was instantiated.
   *
   * This is used in saveChanges, so the entity is only saved to the database if
   * there's any changes.
   * Having this holder array is not as safe as using the loadUnchanged() load
   * function, and only saving if there's any changes.
   *
   * @var array
   * @see saveChanges()
   */
  protected $originalValues = [];

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager = NULL) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->entityTypeManager = $entity_type_manager;

    if ($this->configuration['entity'] instanceof ContentEntityInterface) {
      $this->entity = $this->configuration['entity'];
    }
    else {
      throw new \LogicException('Invalid entity configuration');
    }

    $this->originalValues = $this->entity->toArray();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function createEntity($values = []) {
    $info = entitylogic()->getDefinitionByClass(static::class);

    if (!empty($info['bundle'])) {
      $entity_type = \Drupal::entityTypeManager()->getDefinition($info['entity_type']);
      $values[$entity_type->getKey('bundle')] = $info['bundle'];
    }

    return \Drupal::entityTypeManager()->getStorage($info['entity_type'])->create($values);
  }

  /**
   * Get a basic select query for the entity.
   */
  public static function selectQuery($alias = NULL) {
    $info = entitylogic()->getDefinitionByClass(static::class);
    $entity_type = \Drupal::entityTypeManager()->getDefinition($info['entity_type']);

    if ($alias === NULL) {
      $alias = $info['entity_type'];

      if (!empty($info['bundle'])) {
        $alias .= '__' . $info['bundle'];
      }
    }

    if (!empty($info['bundle'])) {
      $query = \Drupal::database()->select($entity_type->getDataTable(), $alias)
        ->condition($alias . '.' . $entity_type->getKey('bundle'), $info['bundle']);
    }
    else {
      $query = \Drupal::database()->select($entity_type->getDataTable(), $alias);
    }

    $query->addField($alias, $entity_type->getKey('id'), $alias . '__id');

    return $query;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity() {
    return $this->entity;
  }

  /**
   * Check if a anything has changed.
   */
  public function hasChanges() {
    return json_encode($this->originalValues) !== json_encode($this->entity->toArray());
  }

  /**
   * Save the entity.
   */
  public function saveChanges() {
    if ($this->hasChanges() || $this->entity->isNew()) {
      $this->entity->save();
      $this->originalValues = $this->entity->toArray();
    }

    return $this;
  }

}
