<?php

namespace Drupal\entitylogic;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Plugin\Factory\DefaultFactory;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Plugin\DefaultPluginManager;

/**
 * Plugin manager.
 */
class EntityLogicManager extends DefaultPluginManager {

  public const DEFAULT_FALLBACK_ID = '_fallback';

  /**
   * Static cache for logic instances.
   *
   * This way the same logic instance is used everywhere, allowing for states in
   * the objects.
   *
   * @var EntityLogicInterface[]
   *
   * @todo Handle reloads of the entities, so they also update the static cache.
   */
  protected $cache = [];

  /**
   * Construct.
   *
   * @param \Traversable $namespaces
   *   An object that implements \Traversable which contains the root paths
   *   keyed by the corresponding namespace to look for plugin implementations.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   Cache backend instance to use.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler to invoke the alter hook with.
   */
  public function __construct(\Traversable $namespaces, CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler) {
    parent::__construct('Plugin/EntityLogic', $namespaces, $module_handler, 'Drupal\entitylogic\EntityLogicInterface', 'Drupal\entitylogic\Annotation\EntityLogic');
    $this->alterInfo('entitylogic');
    $this->setCacheBackend($cache_backend, 'entitylogic');
  }

  /**
   * Wrap an entity inside an EntityLogic class.
   */
  public function wrapEntity(ContentEntityInterface $entity, $selector = NULL, $configuration = []) {
    $plugin_id = $this->selectPluginId(
      $entity->getEntityTypeId(),
      $entity->getEntityType()->getKey('bundle') ? $entity->bundle() : NULL,
      $selector
    );

    if ($plugin_id === NULL) {
      return NULL;
    }

    // Use the static cache.
    if (!isset($this->cache[$entity->getEntityTypeId()][$entity->id()][$plugin_id])) {
      $this->cache[$entity->getEntityTypeId()][$entity->id()][$plugin_id] = $this->createInstance($plugin_id, ['entity' => $entity] + $configuration);
    }

    return $this->cache[$entity->getEntityTypeId()][$entity->id()][$plugin_id];
  }

  /**
   * Create an entity and wrap it.
   */
  public function createEntity($input, $values = [], $fallback_id = self::DEFAULT_FALLBACK_ID) {
    $entity = $this->getLogicClass($input, $fallback_id)::createEntity($values);

    return $this->wrapEntity($entity);
  }

  /**
   * Get the EntityLogic class definition.
   */
  public function getLogicClass($input, $fallback_id = self::DEFAULT_FALLBACK_ID) {
    $info = $this->decodePluginId($input);

    $plugin_id = $this->selectPluginId($info['entity_type'], $info['bundle'], $info['selector'], $fallback_id);

    if ($plugin_id) {
      $plugin_definition = $this->getDefinition($plugin_id);
      return DefaultFactory::getPluginClass($plugin_id, $plugin_definition, $this->pluginInterface);
    }
  }

  /**
   * Decode an input string.
   *
   * @param string $input
   *   An input string.
   *
   * @return array
   *   String decoded into an array.
   */
  public function decodePluginId($input) {
    $data = ['entity_type' => NULL, 'bundle' => NULL, 'selector' => NULL];

    $input = explode(':', $input);
    if (!empty($input[1])) {
      $data['selector'] = $input[1];
    }

    $input = explode('.', $input[0]);
    if (!empty($input[1])) {
      $data['bundle'] = $input[1];
    }

    $data['entity_type'] = $input[0];

    return $data;
  }

  /**
   * Create a plugin id.
   *
   * @param string $entity_type_id
   *   Entity type ID.
   * @param string $bundle
   *   Bundle.
   * @param string $selector
   *   Selector.
   */
  public function encodePluginId($entity_type_id, $bundle = NULL, $selector = NULL) {
    $id = $entity_type_id;

    if (!empty($bundle)) {
      $id .= '.' . $bundle;
    }

    if (!empty($selector)) {
      $id .= ':' . $selector;
    }

    return $id;
  }

  /**
   * Get the plugin id matching the criteria.
   *
   * @param string $entity_type_id
   *   Entity type ID.
   * @param string $bundle
   *   Bundle.
   * @param string $selector
   *   Selector.
   * @param string $fallback_id
   *   Fallback ID.
   */
  public function selectPluginId($entity_type_id, $bundle = NULL, $selector = NULL, $fallback_id = self::DEFAULT_FALLBACK_ID) {
    $definitions = $this->getDefinitions();

    $plugin_ids = [
      $this->encodePluginId($entity_type_id, $bundle, $selector),
      $this->encodePluginId($entity_type_id, NULL, $selector),
    ];

    if (!empty($fallback_id)) {
      $plugin_ids[] = $this->encodePluginId($fallback_id, NULL, $selector);
    }

    foreach ($plugin_ids as $plugin_id) {
      if (isset($definitions[$plugin_id])) {
        return $plugin_id;
      }
    }
  }

  /**
   * The the definition with the given class name.
   *
   * Include full namespace.
   */
  public function getDefinitionByClass($class_name) {
    foreach ($this->getDefinitions() as $definition) {
      if ($definition['class'] === $class_name) {
        return $definition;
      }
    }
  }

}
